Index: kernel/application.php
===================================================================
--- kernel/application.php	(revision 13862)
+++ kernel/application.php	(working copy)
@@ -2152,7 +2152,13 @@
 
 		if ($this->isDebugMode() && (constOn('DBG_REDIRECT') || (constOn('DBG_RAISE_ON_WARNINGS') && $this->Application->Debugger->WarningCount))) {
 			$this->Debugger->appendTrace();
-			echo "<b>Debug output above!!!</b> Proceed to redirect: <a href=\"$location\">$location</a><br>";
+			echo '<strong>Debug output above !!!</strong><br/>' . "\n";
+
+			if ( array_key_exists('HTTP_REFERER', $_SERVER) ) {
+				echo 'Referer: <strong>' . $_SERVER['HTTP_REFERER'] . '</strong><br/>' . "\n";
+			}
+
+			echo "Proceed to redirect: <a href=\"{$location}\">{$location}</a><br/>\n";
 		}
 		else {
 			if ($js_redirect) {
Index: kernel/utility/http_query.php
===================================================================
--- kernel/utility/http_query.php	(revision 13862)
+++ kernel/utility/http_query.php	(working copy)
@@ -249,11 +249,8 @@
 				$this->processRewriteURL();
 			}
 
-			if (!defined('GW_NOTIFY') && !$rewrite_url && preg_match('/[\/]{0,1}index.php[\/]{0,1}/', $_SERVER['PHP_SELF']) && ($this->Get('t') != 'index')) {
-				// not in payment gateway notification script AND
-				// rewrite url is missing AND not a script from tools folder AND
-				// "index.php" was visited
-				// not on index page
+			if ( !$rewrite_url && $this->rewriteRedirectRequired() ) {
+				// rewrite url is missing (e.g. not a script from tools folder)
 				$url_params = $this->getRedirectParams();
 
 				// no idea about how to check, that given template require category to be passed with it, so pass anyway
@@ -269,6 +266,73 @@
 		}
 	}
 
+	/**
+	 * Checks, that non-rewrite url was visited and it's automatic rewrite is required
+	 *
+	 * @return bool
+	 */
+	function rewriteRedirectRequired()
+	{
+		$redirect_conditions = Array (
+			!$this->Application->Session->IsHTTPSRedirect(), // not https <-> http redirect
+			!$this->refererIsOurSite(), // referer doesn't match ssl path or non-ssl domain (same for site domains)
+			!defined('GW_NOTIFY'), // not in payment gateway notification script
+			preg_match('/[\/]{0,1}index.php[\/]{0,1}/', $_SERVER['PHP_SELF']), // "index.php" was visited
+			$this->Get('t') != 'index', // not on index page
+		);
+
+		$perform_redirect = true;
+
+		foreach ($redirect_conditions as $redirect_condition) {
+			$perform_redirect = $perform_redirect && $redirect_condition;
+
+			if (!$perform_redirect) {
+				return false;
+			}
+		}
+
+		return true;
+	}
+
+	/**
+	 * Checks, that referer is out site
+	 *
+	 * @return bool
+	 */
+	function refererIsOurSite()
+	{
+		if ( !array_key_exists('HTTP_REFERER', $_SERVER) ) {
+			// no referer -> don't care what happens
+			return false;
+		}
+
+		$site_helper =& $this->Application->recallObject('SiteHelper');
+		/* @var $site_helper SiteHelper */
+
+		$found = false;
+		$http_referer = $_SERVER['HTTP_REFERER'];
+		preg_match('/^(.*?):\/\/(.*?)(\/|$)/', $http_referer, $regs); // 1 - protocol, 2 - domain
+
+		if ($regs[1] == 'https') {
+			$found = $site_helper->getDomainByName('SSLUrl', $http_referer) > 0;
+
+			if (!$found) {
+				// check if referer starts with our ssl url
+				$ssl_url = $this->Application->ConfigValue('SSL_URL');
+				$found = $ssl_url && preg_match('/^' . preg_quote($ssl_url, '/') . '/', $http_referer);
+			}
+		}
+		else {
+			$found = $site_helper->getDomainByName('DomainName', $regs[2]) > 0;
+
+			if (!$found) {
+				$found = $regs[2] == DOMAIN;
+			}
+		}
+
+		return $found;
+	}
+
 	function convertFiles()
 	{
 		if (!$_FILES)
Index: units/helpers/site_helper.php
===================================================================
--- units/helpers/site_helper.php	(revision 13862)
+++ units/helpers/site_helper.php	(working copy)
@@ -55,4 +55,91 @@
 
 			return $cache[$iso_code];
 		}
+
+		/**
+		 * Gets site domains from cache
+		 *
+		 * @return Array
+		 */
+		function getSiteDomains()
+		{
+			static $cache = null;
+
+			if (!isset($cache)) {
+				if ($this->Application->isCachingType(CACHING_TYPE_MEMORY)) {
+					$cache = $this->Application->getCache('master:domains_parsed', false);
+				}
+				else {
+					$cache = $this->Application->getDBCache('domains_parsed');
+				}
+
+				if ($cache) {
+					$cache = unserialize($cache);
+				}
+				else {
+					$sql = 'SELECT *
+							FROM ' . TABLE_PREFIX . 'SiteDomains
+							ORDER BY Priority DESC';
+					$cache = $this->Conn->Query($sql, 'DomainId');
+
+					if ($this->Application->isCachingType(CACHING_TYPE_MEMORY)) {
+						$this->Application->setCache('master:domains_parsed', serialize($cache));
+					}
+					else {
+						$this->Application->setDBCache('domains_parsed', serialize($cache));
+					}
+				}
+			}
+
+			return $cache;
+		}
+
+		/**
+		 * Try to match visited domain to any of existing
+		 *
+		 * @param string $field
+		 * @param string $value
+		 * @return int
+		 */
+		function getDomainByName($field, $value)
+		{
+			$site_domains = $this->getSiteDomains();
+			$name_fields = Array ('DomainName', 'SSLUrl');
+
+			foreach ($site_domains as $id => $site_domain) {
+				if ( in_array($field, $name_fields) ) {
+					if (!$site_domain[$field . 'UsesRegExp']) {
+						// not regular expression -> escape manually
+						$site_domain[$field] = preg_quote($site_domain[$field], '/');
+					}
+
+					if (preg_match('/^' . $site_domain[$field] . ($field == 'DomainName' ? '$' : '') . '/', $value)) {
+						return $id;
+					}
+				}
+				elseif ($site_domain[$field] == $value) {
+					return $id;
+				}
+			}
+
+			return false;
+		}
+
+		/**
+		 * Try to match domain settings based on visitor's IP address
+		 *
+		 * @return int
+		 */
+		function getDomainByIP()
+		{
+			$site_domains = $this->getSiteDomains();
+
+			foreach ($site_domains as $id => $site_domain) {
+				if (ipMatch($site_domain['DomainIPRange'], "\n")) {
+					return $id;
+				}
+			}
+
+			return false;
+		}
 	}
Index: units/site_domains/site_domain_eh.php
===================================================================
--- units/site_domains/site_domain_eh.php	(revision 13862)
+++ units/site_domains/site_domain_eh.php	(working copy)
@@ -56,10 +56,12 @@
 
 		function querySiteDomain($field, $value)
 		{
-			$site_domains = $this->getSiteDomains();
+			$site_helper =& $this->Application->recallObject('SiteHelper');
+			/* @var $site_helper SiteHelper */
 
-			$domain_by_name = $this->getDomainByName($field, $value);
-			$domain_by_ip = $this->getDomainByIP();
+			$site_domains = $site_helper->getSiteDomains();
+			$domain_by_name = $site_helper->getDomainByName($field, $value);
+			$domain_by_ip = $site_helper->getDomainByIP();
 
 			if ($domain_by_ip) {
 				$site_domain = $site_domains[$domain_by_ip];
@@ -95,50 +97,6 @@
 		}
 
 		/**
-		 * Try to match visited domain to any of existing
-		 *
-		 * @param string $field
-		 * @param string $value
-		 * @return int
-		 */
-		function getDomainByName($field, $value)
-		{
-			$site_domains = $this->getSiteDomains();
-			$name_fields = Array ('DomainName', 'SSLUrl');
-
-			foreach ($site_domains as $id => $site_domain) {
-				if (in_array($field, $name_fields) && $site_domain[$field . 'UsesRegExp']) {
-					if (preg_match('/^' . $site_domain[$field] . '$/', $value)) {
-						return $id;
-					}
-				}
-				elseif ($site_domain[$field] == $value) {
-					return $id;
-				}
-			}
-
-			return false;
-		}
-
-		/**
-		 * Try to match domain settings based on visitor's IP address
-		 *
-		 * @return int
-		 */
-		function getDomainByIP()
-		{
-			$site_domains = $this->getSiteDomains();
-
-			foreach ($site_domains as $id => $site_domain) {
-				if (ipMatch($site_domain['DomainIPRange'], "\n")) {
-					return $id;
-				}
-			}
-
-			return false;
-		}
-
-		/**
 		 * Load item if id is available
 		 *
 		 * @param kEvent $event
@@ -162,7 +120,10 @@
 				return ;
 			}
 
-			$site_domains = $this->getSiteDomains();
+			$site_helper =& $this->Application->recallObject('SiteHelper');
+			/* @var $site_helper SiteHelper */
+
+			$site_domains = $site_helper->getSiteDomains();
 			$domain_data = array_key_exists($id, $site_domains) ? $site_domains[$id] : false;
 
 			if ($object->LoadFromHash($domain_data)) {
@@ -174,42 +135,6 @@
 			}
 		}
 
-		function getSiteDomains()
-		{
-			static $cache = null;
-
-			if (!isset($cache)) {
-				if ($this->Application->isCachingType(CACHING_TYPE_MEMORY)) {
-					$cache = $this->Application->getCache('master:domains_parsed', false);
-				}
-				else {
-					$cache = $this->Application->getDBCache('domains_parsed');
-				}
-
-				if ($cache) {
-					$cache = unserialize($cache);
-				}
-				else {
-					$id_field = $this->Application->getUnitOption($this->Prefix, 'IDField');
-					$table_name = $this->Application->getUnitOption($this->Prefix, 'TableName');
-
-					$sql = 'SELECT *
-							FROM ' . $table_name . '
-							ORDER BY Priority DESC';
-					$cache = $this->Conn->Query($sql, $id_field);
-
-					if ($this->Application->isCachingType(CACHING_TYPE_MEMORY)) {
-						$this->Application->setCache('master:domains_parsed', serialize($cache));
-					}
-					else {
-						$this->Application->setDBCache('domains_parsed', serialize($cache));
-					}
-				}
-			}
-
-			return $cache;
-		}
-
 		/**
 		 * Removes In-Commerce related fields, when it's not installed
 		 *
